@@ -9,7 +9,7 @@ use 5.006001;
 use strict;
 use warnings;
 
-our $VERSION = '0.10';
+our $VERSION = '0.11';
 
 sub new {
     my ($class, %attr) = @_;
@@ -83,6 +83,9 @@ sub parse_acl {
                 $self->add_alias($k, $v);
             } else {
                 # this is a generic resource
+                unless ($v =~ /^[rw\s]*$/) {
+                    warn "Invalid character in authz rule $v\n";
+                }
                 if (my $resource = $self->resource($current_resource)) {
                     $resource->authorize($k => $v);
                 } else {
@@ -108,8 +111,25 @@ sub verify_acl {
     # Check for references to undefined groups (Thanks Jesse!)
     my (%groups, @errors);
     if ($self->groups) {
+        # gather groups first, in case there are forward refs
         foreach my $group ($self->groups) {
             $groups{$group->name}++;
+            # check for loops
+            local $SIG{__WARN__} = sub { push @errors, @_; };
+            my @g = $self->resolve('@' . $group->name);
+        }
+        foreach my $group ($self->groups) {
+            foreach my $k ($group->members) {
+                if ( $k =~ /^@(.*)/ ) {
+                    unless ( $groups{$1} ) {
+                        push(@errors, "[error] An authz rule (" . $group->name. ") refers to group '$1', which is undefined");
+                    }
+                } elsif ( $k =~ /^&(.*)/ ) {
+                    unless ( $self->aliases->{$1} ) {
+                        push(@errors, "[error] An authz rule (" . $group->name . ") refers to alias '$1', which is undefined");
+                    }
+                }
+            }
         }
     }
 
@@ -120,11 +140,16 @@ sub verify_acl {
                     unless ( $groups{$1} ) {
                         push(@errors, "[error] An authz rule (" . $resource->name . ") refers to group '\@$1', which is undefined");
                     }
+                } elsif ( $k =~ /^&(.*)/ ) {
+                    unless ( $self->aliases->{$1} ) {
+                        push(@errors, "[error] An authz rule (" . $resource->name . ") refers to alias '\&$1', which is undefined");
+                    }
                 }
             }
         }
     }
 
+    chomp @errors;
     return scalar(@errors) ? join("\n", @errors) : undef;
 }
 
@@ -274,14 +299,30 @@ sub add_resource {
         $resource_name = shift(@access);
     }
     
+    my @acl;
+    foreach my $entry (@access) {
+        next if $entry eq "authorized";
+        
+        if (ref($entry) eq "HASH") {
+            # unpack the hashref to a list.
+            foreach my $key (keys %$entry) {
+                push(@acl, $key, $entry->{$key});
+            }
+        } elsif (ref($entry) eq "ARRAY") {
+            push(@acl, @$entry);
+        } else {
+            push(@acl, $entry);
+        }
+    }
+    
     if ($self->resource($resource_name)) {
         die "Can't add new resource $resource_name: resource already exists!\n";
     } elsif ($resource_name !~ /^(?:\S+\:)?\/.*$/) { # Thanks Matt
         die "Invalid resource format in $resource_name! (format 'repo:/path')!\n";
     } else {
         my $resource = SVN::Access::Resource->new(
-            name        =>      $resource_name,
-            authorized  =>      \@access,
+            name => $resource_name,
+            authorized => \@acl,
         );
         push(@{$self->{acl}->{resources}}, $resource);
         return $resource;
@@ -365,11 +406,20 @@ sub group {
 sub resolve {
     my $self = shift;
     my @res;
+    my $seen = (ref $_[$#_] eq "ARRAY" ? pop @_ : []);
 
     foreach my $e (@_) {
         if ($e =~ /^\@(.+)/) {
-            push @res, map $self->resolve($_), $self->group($1)->members()
+            # check for loops
+            if (grep($_ eq $e, @$seen)) {
+                warn "Error: group loop detected ",join(", ", @$seen, $e),"\n";
+                return undef;
+            }
+            push @$seen, $e;
+            push @res, map $self->resolve($_, $seen),
+                           $self->group($1)->members()
                 if $self->group($1);
+            pop @$seen;
         } elsif ($e =~ /^\&(.+)/) {
             push @res, map $self->resolve($_), $self->alias($1)
                 if $self->alias($1);
@@ -598,7 +648,7 @@ Michael Gregorowicz, E<lt>mike@mg2.orgE<gt>
 
 =head1 COPYRIGHT AND LICENSE
 
-Copyright (C) 2012 by Michael Gregorowicz
+Copyright (C) 2012-2017 by Michael Gregorowicz
 
 This library is free software; you can redistribute it and/or modify
 it under the same terms as Perl itself, either Perl version 5.8.8 or,
