@@ -6,7 +6,7 @@ use warnings;
 
 use Tie::IxHash;
 
-our $VERSION = '0.10';
+our $VERSION = '0.11';
 
 sub new {
     my ($class, %attr) = @_;
@@ -15,6 +15,11 @@ sub new {
     my (%authorized, $t);
     $t = tie(%authorized, 'Tie::IxHash');
     
+    # turn hashref into arrayref up front
+    if (ref($attr{authorized}) eq "HASH") {
+        $attr{authorized} = [(%{$attr{authorized}})];
+    }
+    
     # make sure we copy in stuff that was passed.
     %authorized = (@{$attr{authorized}});
     
@@ -151,7 +156,7 @@ Michael Gregorowicz, E<lt>mike@mg2.orgE<gt>
 
 =head1 COPYRIGHT AND LICENSE
 
-Copyright (C) 2012 by Michael Gregorowicz
+Copyright (C) 2012-2017 by Michael Gregorowicz
 
 This library is free software; you can redistribute it and/or modify
 it under the same terms as Perl itself, either Perl version 5.8.8 or,
